/*************************************************************************************/
/*      Copyright 2017 Barcelona Supercomputing Center                               */
/*                                                                                   */
/*      This file is part of the NANOS++ library.                                    */
/*                                                                                   */
/*      NANOS++ is free software: you can redistribute it and/or modify              */
/*      it under the terms of the GNU Lesser General Public License as published by  */
/*      the Free Software Foundation, either version 3 of the License, or            */
/*      (at your option) any later version.                                          */
/*                                                                                   */
/*      NANOS++ is distributed in the hope that it will be useful,                   */
/*      but WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                */
/*      GNU Lesser General Public License for more details.                          */
/*                                                                                   */
/*      You should have received a copy of the GNU Lesser General Public License     */
/*      along with NANOS++.  If not, see <http://www.gnu.org/licenses/>.             */
/*************************************************************************************/

#ifndef POLLINGSERVICE_HPP
#define POLLINGSERVICE_HPP

#include "debug.hpp"
#include "spin_mutex.hpp"

#include "nanos-int.h"

#include "mutex.hpp"

namespace nanos {

class PollingService {
   public:
      PollingService( const char* service_name, nanos_polling_service_t function, void* data ) :
         _function(function),
         _data(data)
      {
      }

      // Calls the polling service function.
      // Returns true when the service have to be automatically unregistered.
      bool operator()() {
         return _function(_data) != 0;
      }

   private:
      nanos_polling_service_t _function;
      void*                   _data;
};

// Naive implementation that only supports a single polling service.
// Aborts if more than a single service is registered.
class PollingServiceManager {
   public:
      static void registerService( const char* name, nanos_polling_service_t function, void* data ) {
         UniqueLock<spin_mutex> guard(_mutex);
         if( !_service ) {
            _service = new PollingService( name, function, data );
         } else {
            fatal( "This polling implementation only supports a single service." );
         }
      }

      static void unregisterService( const char* name, nanos_polling_service_t function, void* data ) {
         UniqueLock<spin_mutex> guard(_mutex);
         unregisterService();
      }

      static void poll() {
         // Call service function.
         UniqueLock<spin_mutex> guard(_mutex, try_to_lock );
         if( guard.owns_lock() && _service != NULL ) {
            int rval = (*_service)();
            if( rval != 0 ) {
               unregisterService();
            }
         }
      }

   private:
      static void unregisterService() {
         delete _service;
         _service = NULL;
      }

      static spin_mutex      _mutex;
      static PollingService* _service;
};

} // namespace nanos

#endif // POLLINGSERVICE_HPP

