/*************************************************************************************/
/*      Copyright 2015 Barcelona Supercomputing Center                               */
/*                                                                                   */
/*      This file is part of the NANOS++ library.                                    */
/*                                                                                   */
/*      NANOS++ is free software: you can redistribute it and/or modify              */
/*      it under the terms of the GNU Lesser General Public License as published by  */
/*      the Free Software Foundation, either version 3 of the License, or            */
/*      (at your option) any later version.                                          */
/*                                                                                   */
/*      NANOS++ is distributed in the hope that it will be useful,                   */
/*      but WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                */
/*      GNU Lesser General Public License for more details.                          */
/*                                                                                   */
/*      You should have received a copy of the GNU Lesser General Public License     */
/*      along with NANOS++.  If not, see <http://www.gnu.org/licenses/>.             */
/*************************************************************************************/

#ifndef _NANOS_LIB_DEBUG
#define _NANOS_LIB_DEBUG

#include <stdexcept>
#include "xstring.hpp"
#include <iostream>

#define _nanos_ostream ( /* myThread ? *(myThread->_file) : */ std::cerr )

namespace nanos {

   inline void fatal_error( const std::string &message, int id )
   {
      _nanos_ostream << "FATAL ERROR: [" << id << "] " << message << std::endl;
      std::terminate();
   }

   inline void fatal_error( const std::string &message )
   {
      _nanos_ostream << "FATAL ERROR: [?] " << message << std::endl;
      std::terminate();
   }

   inline void failed_assertion( const char* file, const int line, const char* value,
                                 const std::string& message, int id )
   {
      _nanos_ostream << "ASSERT failed: [" << id << "] "
                << value << ":" << message
                << " (" << file << ":" << line << ")"
                << std::endl;
      std::terminate();
   }

   inline void failed_assertion( const char* file, int line, const char* value,
                                 const std::string& message )
   {
      _nanos_ostream << "ASSERT failed: [?] "
                << value << ":" << message
                << " (" << file << ":" << line << ")"
                << std::endl;
      std::terminate();
   }

   inline void check_assertion( const char* file, int line, const char* value,
                                bool satisfied, const std::string& message )
   {
      if( !satisfied )
         failed_assertion( file, line, value, message );
   }

   inline void check_assertion( const char* file, int line, const char* value,
                                bool satisfied, const std::string& message, int id )
   {
      if( !satisfied )
         failed_assertion( file, line, value, message, id );
   }


   void printBt( std::ostream &o );

} // namespace nanos

#define fatal(msg)  do { std::stringstream _text; _text << msg; nanos::fatal_error(_text.str()); } while(0)
#define fatal0(msg) do { std::stringstream _text; _text << msg; nanos::fatal_error(_text.str()); } while(0)

#define fatal_cond(cond,msg)  if ( cond ) { std::stringstream _text; _text << msg; fatal_error(_text.str()); }
#define fatal_cond0(cond,msg) if ( cond ) { std::stringstream _text; _text << msg; fatal_error(_text.str()); }

#define warning(msg)  { _nanos_ostream << "WARNING: [" << std::dec << getMyThreadSafe()->getId() << "]" << msg << std::endl; }
#define warning0(msg) { _nanos_ostream << "WARNING: [?]" << msg << std::endl; }

#define message(msg) \
   _nanos_ostream << "MSG: [" << std::dec << getMyThreadSafe()->getId() << "] " << msg << std::endl;
#define message0(msg) \
   _nanos_ostream << "MSG: [?] " << msg << std::endl;

#define messageMaster(msg) \
   if (sys.getNetwork()->getNodeNum() == 0) { _nanos_ostream << "MSG: m:[" << std::dec << getMyThreadSafe()->getId() << "] " << msg << std::endl; }
#define message0Master(msg) \
   if (sys.getNetwork()->getNodeNum() == 0) { _nanos_ostream << "MSG: m:[?] " << msg << std::endl; }

#ifdef NANOS_DEBUG_ENABLED
#define ensure(cond,msg) do { std::stringstream _text; _text << msg; nanos::check_assertion(__FILE__, __LINE__ , #cond, cond, _text.str() ); } while(0)
#define ensure0(cond,msg) do { std::stringstream _text; _text << msg; nanos::check_assertion(__FILE__, __LINE__, #cond, cond, _text.str() ); } while(0)

#define verbose(msg) \
   if (sys.getVerbose()) _nanos_ostream << "[" << std::dec << getMyThreadSafe()->getId() << "]" << msg << std::endl;
#define verbose0(msg) \
   if (sys.getVerbose()) _nanos_ostream << "[?]" << msg << std::endl;

#define debug(msg) \
   if (sys.getVerbose()) _nanos_ostream << "DBG: [" << std::dec << getMyThreadSafe()->getId() << "]" << msg << std::endl;
#define debug0(msg) \
   if (sys.getVerbose()) _nanos_ostream << "DBG: [?]" << msg << std::endl;

#else

#define ensure(cond,msg)
#define ensure0(cond,msg)
#define verbose(msg)
#define verbose0(msg)
#define debug(msg)
#define debug0(msg)

#endif

#endif
